import json
import string
import random
import os
import secrets
import time
from datetime import timedelta, datetime

from django.core.cache import cache
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.http.response import JsonResponse
from jsonschema import validate
from jsonschema.exceptions import ValidationError
from rest_framework import exceptions, generics, permissions, status
from rest_framework.parsers import FileUploadParser
from rest_framework.permissions import IsAuthenticated
from rest_framework.views import APIView

from backend.role_permission import IsProvider
from consents.serializers import (ProviderSchemaSerializer,
                                  ProvidersSerializer,
                                  SaveUserConsentSerializer, CreateUserSerializer, CreateUserShibbolethSerializer)
from providers.models import AnalyticsToken, AnalyticsTokenVerb, Provider, ProviderAuthorization, ProviderSchema
from providers.serializers import (AnalyticsTokenSerializer, ConsentUserVerbThirdPartySerializer,
                                   GetUsersConsentsThirdPartySerializer)
from users.models import CustomUser
from xapi.views import shib_connector_resolver, shib_connector_resolver_to_pairwaise_id

from .models import UserConsents

PROJECT_PATH = os.path.abspath(os.path.dirname(__name__))


class JsonUploadParser(FileUploadParser):
    media_type = "application/json"


def render_provider_consent(provider_schema, supersedes=None):
    return {
        "id": provider_schema.provider.id,
        "name": provider_schema.provider.name,
        "superseded_by": provider_schema.superseded_by.id
        if provider_schema.superseded_by
        else None,
        "definition": provider_schema.definition,
        "createdAt": provider_schema.created,
        "supersedes": render_provider_consent(supersedes) if supersedes else None,
    }


def find_user_verb(verb_id, user_verbs):
    for user_verb in user_verbs:
        if user_verb.get("verb") == verb_id:
            return user_verb

    return None


def merge_consents(provider_schema: ProviderSchema, user_verbs):
    """
    Merges provider schema with actual user consent settings
    """
    user_consent = {
        "id": provider_schema.provider.id,
        "name": provider_schema.provider.name,
        "description": provider_schema.provider.description,
        "groups": provider_schema.groups,
        "essential_verbs": provider_schema.essential_verbs,
    }
    for group in user_consent.get("groups"):
        for verb in group.get("verbs", []):
            user_verb = find_user_verb(verb.get("id"), user_verbs)
            if user_verb:
                verb.update({"consented": user_verb.get("consented")})
                verb.update({"objects": json.loads(user_verb.get("object"))})
    return user_consent


def get_user_consent(user, provider_id):
    """
    Get all user consents for a given provider and merge them in a provider schema.
    """
    user_consents = UserConsents.objects.filter(user=user, provider__pk=provider_id, active=True)

    if user_consents.first():
        provider_schema = user_consents.first().provider_schema
        user_consent_definition = merge_consents(
            provider_schema, user_consents.values()
        )
        return user_consent_definition
    return None


def save_user_consent(user, provider_schema, verbs):
    """
    Inactivates existing user consents for user and provider. Afterwards, new user consents are created.
    """
    old_user_consents = UserConsents.objects.filter(
        user=user, provider=provider_schema.provider.id
    ).update(active=False)

    for verb in verbs:
        try:
            UserConsents.objects.get(user=user, verb=verb["id"], provider=provider_schema.provider.id, active=True)
            # user consent for verb already exists, skip
        except UserConsents.DoesNotExist:
            provider = Provider.objects.get(pk=provider_schema.provider.id)
            UserConsents.objects.create(
                user=user,
                consented=verb.get("consented"),
                provider_schema=provider_schema,
                provider=provider,
                verb=verb.get("id"),
                object=verb.get("objects"),
            )


def is_provider_verb_or_object_missing(
    provider_schema: ProviderSchema, user_consent: dict
) -> bool:
    """
    Validate all provider schema verbs and objects are present in user consent
    """
    provider_schema_verbs = {}

    for group in provider_schema.groups:
        for verb in group["verbs"]:
            provider_schema_verbs[verb["id"]] = [
                object["id"] for object in verb["objects"]
            ]

    for provider_verb_id, provider_object_ids in provider_schema_verbs.items():
        user_verb_consent = next(
            (verb for verb in user_consent["verbs"] if verb["id"] == provider_verb_id),
            None,
        )

        if user_verb_consent is None:
            # Missing provider schema verb id in user consent
            return True

        user_verb_consent_objects_id = [
            object["id"]
            for object in json.loads(user_verb_consent.get("objects", "[]"))
        ]

        if any(
            object_id not in user_verb_consent_objects_id
            for object_id in provider_object_ids
        ):
            # Missing provider schema object id in user consent
            return True

    return False


def validate_distinct_group_ids(provider_schema):
    group_ids = [group["id"] for group in provider_schema["groups"]]
    if len(group_ids) != len(set(group_ids)):
        raise ValueError("Provider schema contains ambiguous group ids.")


def group_consents_by_date(consents):
    """
    Groups consent declaration by provider and date (windows size 1 minute)
    to identify those likely created within the same action.
    """
    # Ensure objects are sorted by their created timestamp
    consents = sorted(consents, key=lambda con: con.created)

    groups = []  # To hold the result
    current_group = {"created": None, "consents": {}}

    for obj in consents:
        if current_group["created"] is None:
            current_group["created"] = obj.created.replace(tzinfo=None)
        if obj.provider.id not in current_group["consents"].keys():
            current_group["consents"][obj.provider.id] = [obj]
        else:
            # Check if the object can be added to the current group
            time_diff = obj.created.replace(tzinfo=None) - current_group["created"]
            if time_diff <= timedelta(minutes=1):
                current_group["consents"][obj.provider.id].append(obj)
                # Recalculate the average timestamp for the current group
                all_current_consents = []
                for _, consents in current_group["consents"].items():
                    all_current_consents += consents
                total_time = sum((o.created.timestamp() for o in all_current_consents), 0)
                avg_time = total_time / sum([len(subgroup) for _, subgroup in current_group["consents"].items()])
                current_group["created"] = datetime.fromtimestamp(avg_time)
            else:
                # Finalize the current group and start a new one
                groups.append(current_group)
                current_group = {"created": obj.created.replace(tzinfo=None), "consents": {obj.provider.id: [obj]}}

    if len(current_group["consents"]) > 0:
        groups.append(current_group)

    return groups

class GetProviderSchemasView(generics.ListAPIView):
    """
    Lists all providers including schema history.
    """

    queryset = Provider.objects.all()
    serializer_class = ProvidersSerializer
    permission_classes = [IsAuthenticated]


class GetProviderStatusThirdPartyView(APIView):
    """
    Allows a third party to query the details of one's provider.
    """

    def get(self, request):
        application_token = request.headers.get("Authorization", "").split("Basic ")[-1]
        provider = ProviderAuthorization.objects.filter(key=application_token).first()
        if provider is None:
            return JsonResponse(
                {"message": "invalid access token"},
                safe=False,
                status=status.HTTP_401_UNAUTHORIZED,
            )
        serializer = ProvidersSerializer(provider.provider)
        return JsonResponse(serializer.data, safe=False, status=status.HTTP_200_OK)


class CreateProviderConsentView(APIView):
    """
    Providers can upload provider schemas. In case a provider schema already exists, the existing provider schema references the new provider schema.
    The latest provider schema has no superseding provider schema associated.
    """

    permission_classes = [IsAuthenticated, IsProvider]
    parser_class = (JsonUploadParser,)

    def put(self, request, format=None):
        file = request.data.get("provider-schema", None)
        if file == None:
            return JsonResponse(
                {"message": "invalid provider schema file"},
                status=status.HTTP_400_BAD_REQUEST,
            )

        # Create or update provider
        try:
            with open(
                os.path.join(PROJECT_PATH, "static/provider_schema.schema.json")
            ) as f:
                provider_schema = json.load(file.file)
                validate(instance=provider_schema, schema=json.load(f))

                validate_distinct_group_ids(provider_schema)

                providerModel = Provider.objects.get(
                    definition_id=provider_schema["id"]
                )  # maybe TODO: match via name, not definition_id, iff definition_id can change each definition
                providerModel.description = provider_schema["description"]
                providerModel.save()
        except OSError:
            return JsonResponse(
                {"message": "Schema missing on server."},
                status=status.HTTP_500_INTERNAL_SERVER_ERROR,
            )
        except ValidationError as e:
            return JsonResponse(
                {"message": "Provider schema validation failed.", "errors": str(e)},
                status=status.HTTP_400_BAD_REQUEST,
            )
        except json.JSONDecodeError as e:
            return JsonResponse(
                {"message": "Decoding json failed.", "errors": str(e)},
                status=status.HTTP_400_BAD_REQUEST,
            )
        except ValueError as e:
            return JsonResponse(
                {"message": str(e)},
                status=status.HTTP_400_BAD_REQUEST,
            )
        except ObjectDoesNotExist:
            providerModel = Provider.objects.create(
                definition_id=provider_schema["id"],
                name=provider_schema["name"],
                description=provider_schema["description"],
            )
            auth_key = secrets.token_hex(
                32
            )  # generate random auth key for new provider

            ProviderAuthorization.objects.create(provider=providerModel, key=auth_key)

        # Create new provider schema and link older schema version to new one (if old one exists)
        try:
            precedingProviderSchema = ProviderSchema.objects.get(
                provider=providerModel, superseded_by__isnull=True
            )

            if precedingProviderSchema:
                providerSchemaModel = ProviderSchema.objects.create(
                    provider=providerModel,
                    groups=provider_schema["groups"],
                    essential_verbs=provider_schema["essentialVerbs"],
                    additional_lrs=provider_schema["additionalLrs"] if "additionalLrs" in provider_schema.keys() else []
                )
                precedingProviderSchema.superseded_by = providerSchemaModel
                precedingProviderSchema.save()

                provider = precedingProviderSchema.provider
                provider.description = provider_schema["description"]
                provider.name=provider_schema["name"]
                provider.save()

        except ObjectDoesNotExist:
            providerSchemaModel = ProviderSchema.objects.create(
                provider=providerModel,
                groups=provider_schema["groups"],
                essential_verbs=provider_schema["essentialVerbs"],
                additional_lrs=provider_schema["additionalLrs"] if "additionalLrs" in provider_schema.keys() else []
            )

        return JsonResponse(
            {"message": "provider schema processed"}, status=status.HTTP_201_CREATED
        )


class GetUserConsentView(APIView):
    """
    Users are able to make a consent decision for a specific provider schema (always the latest version at time).
    This endpoint returns the user consent along with the associated provider schema, which might not be the latest version anymore.
    """

    permission_classes = (IsAuthenticated,)

    def get(self, request, provider_id):
        user = request.user
        try:
            user_consent = get_user_consent(user, provider_id)

            if user_consent is None:
                return JsonResponse(
                    {
                        "consent": None,
                        "paused_data_recording": user.paused_data_recording,
                    },
                    safe=False,
                    status=status.HTTP_200_OK,
                )
            else:
                accepted_provider_schemas = user.accepted_provider_schemas.all()
                provider_schemas = list(
                    filter(
                        lambda schema: schema.provider.id == int(provider_id),
                        accepted_provider_schemas,
                    )
                )

                provider_schema = (
                    ProviderSchemaSerializer(provider_schemas, many=True).data[0]
                    if len(provider_schemas) > 0
                    else None
                )
                return JsonResponse(
                    {
                        "consent": user_consent,
                        "provider_schema": provider_schema,
                        "paused_data_recording": user.paused_data_recording,
                    },
                    safe=False,
                    status=status.HTTP_200_OK,
                )
        except ObjectDoesNotExist:
            return JsonResponse(
                {
                    "message": "User has no consent declaration record.",
                    "no_consent_record": True,
                },
                safe=False,
                status=status.HTTP_400_BAD_REQUEST,
            )


class GetUserConsentHistoryView(APIView):
    """
    This endpoint returns all consents declared by the user, sorted by creation date.
    """

    permission_classes = (IsAuthenticated,)

    def get(self, request):
        user = request.user
        try:
            user_consents = UserConsents.objects.filter(user=user).order_by('created')

            if not user_consents.first():
                return JsonResponse(
                    {
                        "consents": [],
                    },
                    safe=False,
                    status=status.HTTP_200_OK,
                )
            else:
                consents_grouped = group_consents_by_date(consents=user_consents)
                consents_mapped = []
                for consent_group in consents_grouped:
                    consent_group_mapped = {
                        "created": consent_group["created"],
                        "consents": []
                    }
                    for provider_id, consents in consent_group["consents"].items():
                        provider_schema = consents[0].provider_schema
                        user_consent = {
                            "id": provider_schema.provider.id,
                            "name": provider_schema.provider.name,
                            "description": provider_schema.provider.description,
                            "groups": provider_schema.groups,
                            "essential_verbs": provider_schema.essential_verbs,
                            "created": consent_group["created"],
                        }
                        for group in user_consent.get("groups"):
                            for verb in group.get("verbs", []):
                                user_verb = find_user_verb(verb.get("id"), [consent.__dict__ for consent in consents])
                                if user_verb:
                                    verb.update({"consented": user_verb.get("consented")})
                                    verb.update({"objects": json.loads(user_verb.get("object"))})
                        consent_group_mapped["consents"].append(user_consent)
                    consents_mapped.append(consent_group_mapped)

                return JsonResponse(
                    {
                        "groups": consents_mapped
                    },
                    safe=False,
                    status=status.HTTP_200_OK,
                )
        except ObjectDoesNotExist:
            return JsonResponse(
                {
                    "message": "User has no consent declaration record.",
                    "no_consent_record": True,
                },
                safe=False,
                status=status.HTTP_400_BAD_REQUEST,
            )

class CreateUserConsentView(APIView):
    
    def post(self, request):
        application_token = request.headers.get("Authorization", "").split("Basic ")[-1]
        provider = ProviderAuthorization.objects.filter(key=application_token).first()
        if provider is None:
            return JsonResponse(
                {"message": "invalid access token"},
                safe=False,
                status=status.HTTP_401_UNAUTHORIZED,
            )
       
        serializer = CreateUserSerializer(data=request.data)
        serializer.is_valid(raise_exception=True) 
        
        email = serializer.validated_data["email"]

        if CustomUser.objects.filter(email=email).first() is None:
           user = CustomUser.objects.create(
                email=email,
                first_name=serializer.validated_data["first_name"],
                last_name=serializer.validated_data["last_name"]
           )

        return JsonResponse(
                    {
                        "user": "user is created",
                    },
                    safe=False,
                    status=status.HTTP_200_OK,
                )
    
class CreateUserConsentViaConnectServiceView(APIView):
    
    def post(self, request):
        application_token = request.headers.get("Authorization", "").split("Basic ")[-1]
        provider = ProviderAuthorization.objects.filter(key=application_token).first()
        if provider is None:
            return JsonResponse(
                {"message": "invalid access token"},
                safe=False,
                status=status.HTTP_401_UNAUTHORIZED,
            )
       
        serializer = CreateUserShibbolethSerializer(data=request.data)
        serializer.is_valid(raise_exception=True) 
        
        email = serializer.validated_data["email"]
        shib_id = shib_connector_resolver_to_pairwaise_id(email=email, provider=provider)

        user = CustomUser.objects.filter(shibboleth_connector_identifier=shib_id).first()

        message = "user exists"
        if not user:
            lock_key = f"user_creation_lock_{shib_id}"  # "mutex" to prevent race conditions using a cache entry
            if cache.add(lock_key, "locked", 2):  # 2 is a timeout value, after wich the cache entry is deleted
                try:
                    user = CustomUser.objects.create(
                        shibboleth_connector_identifier=shib_id, # this is the pairwise id
                        email=''.join(random.choices(string.ascii_uppercase + string.digits, k=8)) + "@manual-created.polaris",
                        first_name=''.join(random.choices(string.ascii_uppercase + string.digits, k=8)),
                        last_name=''.join(random.choices(string.ascii_uppercase + string.digits, k=8))
                    )
                except Exception as e:
                    return JsonResponse({"message": "user could not be created"}, safe=False,
                                        status=status.HTTP_500_INTERNAL_SERVER_ERROR)
                finally:
                    message = "user has been created"
                    cache.delete(lock_key)
            else:
                for _ in range(4):  # Wait up to 2x lock_timeout
                    time.sleep(0.1)
                    if not cache.get(lock_key):
                        break
                # Fetch the user created by another process
                user = CustomUser.objects.filter(shibboleth_connector_identifier=shib_id).first()
                if not user:
                    return JsonResponse({"message": "user could not be created"}, safe=False,
                                        status=status.HTTP_500_INTERNAL_SERVER_ERROR)

        return JsonResponse(
                    {
                        "message": message,
                    },
                    safe=False,
                    status=status.HTTP_200_OK,
                )

class SaveUserConsentView(APIView):
    """
    Saves a user consent.
    The provider schema id for each provider must always refer to the latest schema version of the respective provider, otherwise the request will be rejected
    """

    permission_classes = (IsAuthenticated,)

    def post(self, request):
        user = request.user

        serializer = SaveUserConsentSerializer(data=request.data, many=True)
        serializer.is_valid(raise_exception=True)

        # Validate each user provider setting refers to the latest provider schema id
        for user_provider_setting in serializer.validated_data:
            latest_provider_schema = ProviderSchema.objects.get(
                provider=user_provider_setting["providerId"].id, superseded_by=None
            )

            if (
                user_provider_setting["providerSchemaId"].id
                != latest_provider_schema.id
            ):
                return JsonResponse(
                    {"message": "provider schema id isn't latest provider schema id"},
                    status=status.HTTP_400_BAD_REQUEST,
                )

            if is_provider_verb_or_object_missing(
                latest_provider_schema, user_provider_setting
            ):
                return JsonResponse(
                    {
                        "message": "user consent contains missing provider schema verb consents"
                    },
                    safe=False,
                    status=status.HTTP_400_BAD_REQUEST,
                )

        for user_provider_setting in serializer.validated_data:
            user.accepted_provider_schemas.add(
                user_provider_setting["providerSchemaId"]
            )
            user.save()
            save_user_consent(
                user,
                user_provider_setting["providerSchemaId"],
                user_provider_setting["verbs"],
            )

        return JsonResponse(
            {"message": "user consent saved"},
            status=status.HTTP_200_OK,
        )

class SaveUserConsentAnalyticsTokens(APIView):
    permission_classes = (IsAuthenticated,)

    def post(self, request):
        user = request.user
        token = AnalyticsToken.objects.filter(id = request.data["analyticTokenId"]).first()
        if token is None:
            return JsonResponse(
                {"message": "not found"},
                status=status.HTTP_400_BAD_REQUEST,
            )
        verbs = AnalyticsTokenVerb.objects.filter(analytics_token_id = token.id)
        for verb in verbs:
            consent = UserConsents.objects.filter(verb = verb.verb, user_id=user.id, provider_id = verb.provider, active=True).first()
            consent.consented = True
            consent.save()

        return JsonResponse(
            {"message": "user consent saved"},
            status=status.HTTP_200_OK,
        )

class GetUserConsentAnalyticsTokens(generics.ListAPIView):
    """
    Lists analytics tokens for the user
    """
    def get_queryset(self):
        return AnalyticsToken.objects.filter()

   # queryset = AnalyticsToken.objects.filter(creator=request.user)
    serializer_class = AnalyticsTokenSerializer
    permission_classes = [IsAuthenticated]

class GetUserConsentProviders(generics.ListAPIView):
    """
    Lists providers for the user
    """
    def get_queryset(self):
        return Provider.objects.filter()

   # queryset = AnalyticsToken.objects.filter(creator=request.user)
    serializer_class = ProvidersSerializer
    permission_classes = [IsAuthenticated]


class GetUserConsentStatusView(APIView):
    """
    Checks whether user accepted the latest provider schema for each provider, an old consent declaration
    or none and returns "latest", "outdated" or "none" respectively.
    """

    permission_classes = (IsAuthenticated,)

    def get(self, request):
        providers = Provider.objects.all()
        user = request.user

        result = {}
        accepted_user_schemas = user.accepted_provider_schemas.all()
        user_schema_ids = [schema.id for schema in accepted_user_schemas]

        for provider in providers:
            provider_latest_schema = ProviderSchema.objects.get(
                provider=provider, superseded_by=None
            )

            if len(accepted_user_schemas) == 0:
                result[provider.id] = {"status": "none"}
            else:
                acceptance_status = (
                    "latest"
                    if provider_latest_schema.id in user_schema_ids
                    else "outdated"
                )
                result[provider.id] = {
                    "status": acceptance_status,
                    "userConsent": get_user_consent(user, provider.id),
                }
        return JsonResponse({"result": result}, safe=False, status=status.HTTP_200_OK)


class GetUsersConsentsThirdPartyView(APIView):
    """
    Obtain user consent for a single user.
    This endpoint should only be used by a third party.
    """

    def get(self, request, user_id):
        application_token = request.headers.get("Authorization", "").split("Basic ")[-1]
        provider = ProviderAuthorization.objects.filter(key=application_token).first()
        if provider is None:
            return JsonResponse(
                {"message": "invalid access token"},
                safe=False,
                status=status.HTTP_401_UNAUTHORIZED,
            )
        
        email = user_id      
        email = shib_connector_resolver(email=email, provider=provider)
        serializer = GetUsersConsentsThirdPartySerializer(data={"user": email})
        serializer.is_valid(raise_exception=True)

        user = CustomUser.objects.filter(email=email).first()
        try:
            user_consent = get_user_consent(user, provider.provider.id)

            if user_consent is None:
                return JsonResponse(
                    {
                        "consent": None,
                        "paused_data_recording": user.paused_data_recording,
                    },
                    safe=False,
                    status=status.HTTP_200_OK,
                )
            else:
                accepted_provider_schemas = user.accepted_provider_schemas.all()
                provider_schemas = list(
                    filter(
                        lambda schema: schema.provider.id == int(provider.provider.id),
                        accepted_provider_schemas,
                    )
                )

                provider_schema = (
                    ProviderSchemaSerializer(provider_schemas, many=True).data[0]
                    if len(provider_schemas) > 0
                    else None
                )
                return JsonResponse(
                    {
                        "consent": user_consent,
                        "provider_schema": provider_schema,
                        "paused_data_recording": user.paused_data_recording,
                    },
                    safe=False,
                    status=status.HTTP_200_OK,
                )
        except ObjectDoesNotExist:
            return JsonResponse(
                {
                    "message": "User has no consent declaration record.",
                    "no_consent_record": True,
                },
                safe=False,
                status=status.HTTP_400_BAD_REQUEST,
            )


class SaveUserConsentThirdPartyView(APIView):
    """
    Saves a user consent, with authentication taking place via application tokens.
    The provider schema id for each provider must always refer to the latest schema version of the respective provider, otherwise the request will be rejected
    """

    def post(self, request):
        application_token = request.headers.get("Authorization", "").split("Basic ")[-1]
        provider = ProviderAuthorization.objects.filter(key=application_token).first()
        if provider is None:
            return JsonResponse(
                {"message": "invalid access token"},
                safe=False,
                status=status.HTTP_401_UNAUTHORIZED,
            )

        serializer = ConsentUserVerbThirdPartySerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        email = serializer.validated_data["user_id"]        
        email = shib_connector_resolver(email=email, provider=provider)
        user = CustomUser.objects.filter(email=email).first()

        for verb in serializer.validated_data["verbs"]:
            if verb["provider"].id != provider.provider.id:
                return JsonResponse(
                    {
                        "message": "you are not authorized to access providers other than your own"
                    },
                    safe=False,
                    status=status.HTTP_403_FORBIDDEN,
                )

        # Validate user provider setting refers to the latest provider schema id
        latest_provider_schema = ProviderSchema.objects.get(
            provider=provider.provider, superseded_by=None
        )

        if (
            serializer.validated_data["provider_schema_id"].id
            != latest_provider_schema.id
        ):
            return JsonResponse(
                {"message": "provider schema id isn't latest provider schema id"},
                status=status.HTTP_400_BAD_REQUEST,
            )

        if is_provider_verb_or_object_missing(
            latest_provider_schema, serializer.validated_data
        ):
            return JsonResponse(
                {
                    "message": "user consent contains missing provider schema verb consents"
                },
                safe=False,
                status=status.HTTP_400_BAD_REQUEST,
            )

        save_user_consent(
            user,
            serializer.validated_data["provider_schema_id"],
            serializer.validated_data["verbs"],
        )

        # Add provider schema id to accepted ones for user
        user.accepted_provider_schemas.add(
            serializer.validated_data["provider_schema_id"]
        )
        user.save()
        return JsonResponse(
            {"message": "user consent saved"}, status=status.HTTP_200_OK
        )
